@@ -24,6 +24,7 @@ DROP TABLE IF EXISTS g_client;
 DROP TABLE IF EXISTS g_user_property;
 DROP TABLE IF EXISTS g_user_scope_user;
 DROP TABLE IF EXISTS g_user_scope;
+DROP TABLE IF EXISTS g_user_password;
 DROP TABLE IF EXISTS g_user;
 DROP TABLE IF EXISTS gpg_device_authorization_scope;
 DROP TABLE IF EXISTS gpg_device_authorization;
@@ -32,7 +33,11 @@ DROP TABLE IF EXISTS gpg_access_token;
 DROP TABLE IF EXISTS gpg_refresh_token_scope;
 DROP TABLE IF EXISTS gpg_refresh_token;
 DROP TABLE IF EXISTS gpg_code_scope;
+DROP TABLE IF EXISTS gpo_par_scope;
+DROP TABLE IF EXISTS gpo_par;
+DROP TABLE IF EXISTS gpo_rar;
 DROP TABLE IF EXISTS gpg_code;
+DROP TABLE IF EXISTS gpo_dpop;
 DROP TABLE IF EXISTS gpo_device_scheme;
 DROP TABLE IF EXISTS gpo_device_authorization_scope;
 DROP TABLE IF EXISTS gpo_device_authorization;
@@ -68,6 +73,7 @@ CREATE TABLE g_user_module_instance (
   gumi_display_name VARCHAR(256) DEFAULT '',
   gumi_parameters MEDIUMBLOB,
   gumi_readonly TINYINT(1) DEFAULT 0,
+  gumi_multiple_passwords TINYINT(1) DEFAULT 0,
   gumi_enabled TINYINT(1) DEFAULT 1
 );
 
@@ -221,7 +227,6 @@ CREATE TABLE g_user (
   gu_username VARCHAR(128) NOT NULL UNIQUE,
   gu_name VARCHAR(256) DEFAULT '',
   gu_email VARCHAR(512) DEFAULT '',
-  gu_password VARCHAR(256),
   gu_enabled TINYINT(1) DEFAULT 1
 );
 
@@ -249,6 +254,13 @@ CREATE TABLE g_user_property (
 );
 CREATE INDEX i_g_user_property_name ON g_user_property(gup_name);
 
+CREATE TABLE g_user_password (
+  guw_id INT(11) PRIMARY KEY AUTO_INCREMENT,
+  gu_id INT(11),
+  guw_password VARCHAR(256),
+  FOREIGN KEY(gu_id) REFERENCES g_user(gu_id) ON DELETE CASCADE
+);
+
 CREATE TABLE gpg_code (
   gpgc_id INT(11) PRIMARY KEY AUTO_INCREMENT,
   gpgc_plugin_name VARCHAR(256) NOT NULL,
@@ -357,7 +369,9 @@ CREATE TABLE gpo_code (
   gpoc_redirect_uri VARCHAR(512) NOT NULL,
   gpoc_code_hash VARCHAR(512) NOT NULL,
   gpoc_nonce VARCHAR(512),
+  gpoc_resource VARCHAR(512),
   gpoc_claims_request BLOB DEFAULT NULL,
+  gpoc_authorization_details BLOB DEFAULT NULL,
   gpoc_expires_at TIMESTAMP NOT NULL DEFAULT CURRENT_TIMESTAMP,
   gpoc_issued_for VARCHAR(256), -- IP address or hostname
   gpoc_user_agent VARCHAR(256),
@@ -388,7 +402,9 @@ CREATE TABLE gpo_refresh_token (
   gpoc_id INT(11) DEFAULT NULL,
   gpor_username VARCHAR(256) NOT NULL,
   gpor_client_id VARCHAR(256),
+  gpor_resource VARCHAR(512),
   gpor_claims_request BLOB DEFAULT NULL,
+  gpor_authorization_details BLOB DEFAULT NULL,
   gpor_issued_at TIMESTAMP NOT NULL DEFAULT CURRENT_TIMESTAMP,
   gpor_expires_at TIMESTAMP NOT NULL DEFAULT CURRENT_TIMESTAMP,
   gpor_last_seen TIMESTAMP NOT NULL DEFAULT CURRENT_TIMESTAMP,
@@ -398,6 +414,7 @@ CREATE TABLE gpo_refresh_token (
   gpor_user_agent VARCHAR(256),
   gpor_token_hash VARCHAR(512) NOT NULL,
   gpor_jti VARCHAR(128),
+  gpor_dpop_jkt VARCHAR(512),
   gpor_enabled TINYINT(1) DEFAULT 1,
   FOREIGN KEY(gpoc_id) REFERENCES gpo_code(gpoc_id) ON DELETE CASCADE
 );
@@ -419,11 +436,13 @@ CREATE TABLE gpo_access_token (
   gpor_id INT(11) DEFAULT NULL,
   gpoa_username VARCHAR(256),
   gpoa_client_id VARCHAR(256),
+  gpoa_resource VARCHAR(512),
   gpoa_issued_at TIMESTAMP DEFAULT CURRENT_TIMESTAMP,
   gpoa_issued_for VARCHAR(256), -- IP address or hostname
   gpoa_user_agent VARCHAR(256),
   gpoa_token_hash VARCHAR(512) NOT NULL,
   gpoa_jti VARCHAR(128),
+  gpoa_authorization_details BLOB DEFAULT NULL,
   gpoa_enabled TINYINT(1) DEFAULT 1,
   FOREIGN KEY(gpor_id) REFERENCES gpo_refresh_token(gpor_id) ON DELETE CASCADE
 );
@@ -492,6 +511,7 @@ CREATE TABLE gpo_device_authorization (
   gpoda_id INT(11) PRIMARY KEY AUTO_INCREMENT,
   gpoda_plugin_name VARCHAR(256) NOT NULL,
   gpoda_client_id VARCHAR(256) NOT NULL,
+  gpoda_resource VARCHAR(512),
   gpoda_username VARCHAR(256),
   gpoda_created_at TIMESTAMP DEFAULT CURRENT_TIMESTAMP,
   gpoda_expires_at TIMESTAMP DEFAULT CURRENT_TIMESTAMP,
@@ -499,6 +519,7 @@ CREATE TABLE gpo_device_authorization (
   gpoda_device_code_hash VARCHAR(512) NOT NULL,
   gpoda_user_code_hash VARCHAR(512) NOT NULL,
   gpoda_status TINYINT(1) DEFAULT 0, -- 0: created, 1: user verified, 2 device completed, 3 disabled
+  gpoda_authorization_details BLOB DEFAULT NULL,
   gpoda_last_check TIMESTAMP DEFAULT CURRENT_TIMESTAMP
 );
 CREATE INDEX i_gpoda_device_code_hash ON gpo_device_authorization(gpoda_device_code_hash);
@@ -519,6 +540,62 @@ CREATE TABLE gpo_device_scheme (
   FOREIGN KEY(gpoda_id) REFERENCES gpo_device_authorization(gpoda_id) ON DELETE CASCADE
 );
 
+CREATE TABLE gpo_dpop (
+  gpod_id INT(11) PRIMARY KEY AUTO_INCREMENT,
+  gpod_plugin_name VARCHAR(256) NOT NULL,
+  gpod_client_id VARCHAR(256) NOT NULL,
+  gpod_jti_hash VARCHAR(512) NOT NULL,
+  gpod_jkt VARCHAR(512) NOT NULL,
+  gpod_htm VARCHAR(128) NOT NULL,
+  gpod_htu VARCHAR(512) NOT NULL,
+  gpod_iat TIMESTAMP NOT NULL,
+  gpod_last_seen TIMESTAMP DEFAULT CURRENT_TIMESTAMP
+);
+CREATE INDEX i_gpod_jti_hash ON gpo_dpop(gpod_jti_hash);
+
+CREATE TABLE gpo_rar (
+  gporar_id INT(11) PRIMARY KEY AUTO_INCREMENT,
+  gporar_plugin_name VARCHAR(256) NOT NULL,
+  gporar_client_id VARCHAR(256) NOT NULL,
+  gporar_type VARCHAR(256) NOT NULL,
+  gporar_username VARCHAR(256),
+  gporar_consent TINYINT(1) DEFAULT 0,
+  gporar_enabled TINYINT(1) DEFAULT 1
+);
+CREATE INDEX i_gporar_client_id ON gpo_rar(gporar_client_id);
+CREATE INDEX i_gporar_type ON gpo_rar(gporar_type);
+CREATE INDEX i_gporar_username ON gpo_rar(gporar_username);
+
+CREATE TABLE gpo_par (
+  gpop_id INT(11) PRIMARY KEY AUTO_INCREMENT,
+  gpop_plugin_name VARCHAR(256) NOT NULL,
+  gpop_response_type VARCHAR(128) NOT NULL,
+  gpop_state BLOB,
+  gpop_username VARCHAR(256),
+  gpop_client_id VARCHAR(256) NOT NULL,
+  gpop_redirect_uri VARCHAR(512) NOT NULL,
+  gpop_request_uri_hash VARCHAR(512) NOT NULL,
+  gpop_nonce VARCHAR(512),
+  gpop_code_challenge VARCHAR(128),
+  gpop_resource VARCHAR(512),
+  gpop_claims_request BLOB DEFAULT NULL,
+  gpop_authorization_details BLOB DEFAULT NULL,
+  gpop_additional_parameters BLOB DEFAULT NULL,
+  gpop_status TINYINT(1) DEFAULT 0, -- 0 created, 1 validated, 2 completed
+  gpop_expires_at TIMESTAMP NOT NULL DEFAULT CURRENT_TIMESTAMP,
+  gpop_issued_for VARCHAR(256), -- IP address or hostname
+  gpop_user_agent VARCHAR(256)
+);
+CREATE INDEX i_gpop_request_uri_hash ON gpo_par(gpop_request_uri_hash);
+CREATE INDEX i_gpop_code_challenge ON gpo_par(gpop_code_challenge);
+
+CREATE TABLE gpo_par_scope (
+  gpops_id INT(11) PRIMARY KEY AUTO_INCREMENT,
+  gpop_id INT(11),
+  gpops_scope VARCHAR(128) NOT NULL,
+  FOREIGN KEY(gpop_id) REFERENCES gpo_par(gpop_id) ON DELETE CASCADE
+);
+
 CREATE TABLE gs_code (
   gsc_id INT(11) PRIMARY KEY AUTO_INCREMENT,
   gsc_mod_name VARCHAR(128) NOT NULL,
@@ -684,8 +761,9 @@ INSERT INTO g_scope (gs_name, gs_display_name, gs_description, gs_password_requi
 INSERT INTO g_scope (gs_name, gs_display_name, gs_description, gs_password_required, gs_password_max_age) VALUES ('g_profile', 'Glewlwyd profile', 'Access to the user''s profile API', 1, 600);
 INSERT INTO g_scope (gs_name, gs_display_name, gs_description, gs_password_required, gs_password_max_age) VALUES ('openid', 'Open ID', 'Open ID Connect scope', 0, 0);
 INSERT INTO g_user_module_instance (gumi_module, gumi_order, gumi_name, gumi_display_name, gumi_parameters, gumi_readonly) VALUES ('database', 0, 'database', 'Database backend', '{"use-glewlwyd-connection":true,"data-format":{"picture":{"multiple":false,"read":true,"write":true,"profile-read":true,"profile-write":true},"reset-credentials-code":{"multiple":false,"read":true,"write":true,"profile-read":false,"profile-write":false}}}', 0);
-INSERT INTO g_client_module_instance (gcmi_module, gcmi_order, gcmi_name, gcmi_display_name, gcmi_parameters, gcmi_readonly) VALUES ('database', 0, 'database', 'Database backend', '{"use-glewlwyd-connection":true,"data-format":{"redirect_uri":{"multiple":true,"read":true,"write":true},"authorization_type":{"multiple":true,"read":true,"write":true},"client_secret":{"read":true,"write":true},"pubkey":{"read":true,"write":true},"jwks":{"convert":"jwks","read":true,"write":true},"jwks_uri":{"read":true,"write":true},"post_logout_redirect_uris":{"multiple":true,"read":true,"write":true},"token_endpoint_auth_method":{"multiple":false,"read":true,"write":true}}}', 0);
-INSERT INTO g_user (gu_username, gu_name, gu_password, gu_email, gu_enabled) VALUES ('admin', 'The Administrator', PASSWORD('password'), '', 1);
+INSERT INTO g_client_module_instance (gcmi_module, gcmi_order, gcmi_name, gcmi_display_name, gcmi_parameters, gcmi_readonly) VALUES ('database', 0, 'database', 'Database backend', '{"use-glewlwyd-connection":true,"data-format":{"redirect_uri":{"multiple":true,"read":true,"write":true},"authorization_type":{"multiple":true,"read":true,"write":true},"client_secret":{"read":true,"write":true},"pubkey":{"read":true,"write":true},"jwks":{"convert":"jwks","read":true,"write":true},"jwks_uri":{"read":true,"write":true},"post_logout_redirect_uris":{"multiple":true,"read":true,"write":true},"token_endpoint_auth_method":{"multiple":false,"read":true,"write":true},"resource":{"multiple":true,"read":true,"write":true},"authorization_data_types":{"multiple":true,"read":true,"write":true}}}', 0);
+INSERT INTO g_user (gu_username, gu_name, gu_email, gu_enabled) VALUES ('admin', 'The Administrator', '', 1);
+INSERT INTO g_user_password (gu_id, guw_password) VALUES ((SELECT gu_id from g_user WHERE gu_username='admin'), PASSWORD('password'));
 INSERT INTO g_user_scope (gus_name) VALUES ('g_admin');
 INSERT INTO g_user_scope (gus_name) VALUES ('g_profile');
 INSERT INTO g_user_scope_user (gu_id, gus_id) VALUES ((SELECT gu_id from g_user WHERE gu_username='admin'), (SELECT gus_id FROM g_user_scope WHERE gus_name='g_admin'));
